# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from openlineage.common.provider.great_expectations.action import (
    OpenLineageValidationAction,
)

__all__ = [OpenLineageValidationAction]



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import Dict, List, Optional, Union

import attr
from openlineage.client.facet import BaseFacet
from openlineage.common.provider.great_expectations.results import (
    GreatExpectationsAssertion,
)

from great_expectations.core import IDDict
from great_expectations.core.batch import BatchDefinition, BatchMarkers
from great_expectations.core.id_dict import BatchKwargs, BatchSpec


@attr.s
class GreatExpectationsRunFacet(BaseFacet):
    """
    Custom facet which describes the instance of GreatExpectations and the suite configuration
    """

    great_expectations_version = attr.ib()
    expectation_suite_name: str = attr.ib()
    run_id: Dict = attr.ib()  # type: ignore
    expectation_suite_meta: Dict = attr.ib()
    validation_time: str = attr.ib()
    batch_spec: Optional[BatchSpec] = attr.ib(default=None)
    batch_markers: Optional[BatchMarkers] = attr.ib(default=None)
    batch_kwargs: Optional[BatchKwargs] = attr.ib(default=None)
    active_batch_definition: Union[None, IDDict, BatchDefinition] = attr.ib(default=None)
    batch_parameters = attr.ib(default=None)
    checkpoint_name: Optional[str] = attr.ib(default=None)
    validation_id: Optional[str] = attr.ib(default=None)
    checkpoint_id: Optional[str] = attr.ib(default=None)

    @staticmethod
    def _get_schema() -> str:
        return "https://github.com/OpenLineage/OpenLineage/tree/main/integration/common/openlineage/common/provider/ge-run-facet.json"


@attr.s
class GreatExpectationsAssertionsDatasetFacet(BaseFacet):
    """
    This facet represents passed/failed status of asserted expectations on dataset
    """

    assertions: List[GreatExpectationsAssertion] = attr.ib()

    @staticmethod
    def _get_schema() -> str:
        return "https://github.com/OpenLineage/OpenLineage/tree/main/integration/common/openlineage/common/provider/ge-assertions-dataset-facet.json"



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from openlineage.common.utils import get_from_nullable_chain, parse_single_arg


def test_nullable_chain_fails():
    x = {"first": {"second": {}}}
    assert get_from_nullable_chain(x, ["first", "second", "third"]) is None


def test_nullable_chain_works():
    x = {"first": {"second": {"third": 42}}}
    assert get_from_nullable_chain(x, ["first", "second", "third"]) == 42

    x = {"first": {"second": {"third": 42, "fourth": {"empty": 56}}}}
    assert get_from_nullable_chain(x, ["first", "second", "third"]) == 42


def test_parse_single_arg_does_not_exist():
    assert parse_single_arg(["dbt", "run"], ["-t", "--target"]) is None
    assert parse_single_arg(["python", "main.py", "--random_arg", "yes"], ["--what"]) is None


def test_parse_single_arg_next():
    assert parse_single_arg(["dbt", "run", "--target", "prod"], ["-t", "--target"]) == "prod"
    assert parse_single_arg(["python", "--random=yes", "--what", "asdf"], ["--what"]) == "asdf"


def test_parse_single_arg_equals():
    assert parse_single_arg(["dbt", "run", "--target=prod"], ["-t", "--target"]) == "prod"
    assert parse_single_arg(["python", "--random=yes", "--what=asdf"], ["--what"]) == "asdf"


def test_parse_single_arg_gets_first_key():
    assert parse_single_arg(["dbt", "run", "--target=prod", "-t=a"], ["-t", "--target"]) == "a"


def test_parse_single_arg_default():
    assert parse_single_arg(["dbt", "run"], ["-t", "--target"]) is None
    assert parse_single_arg(["dbt", "run"], ["-t", "--target"], default="prod") == "prod"



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from openlineage.common.provider.dbt.cloud import DbtCloudArtifactProcessor
from openlineage.common.provider.dbt.local import DbtLocalArtifactProcessor
from openlineage.common.provider.dbt.processor import (
    ParentRunMetadata,
    UnsupportedDbtCommand,
)

__all__ = [
    DbtLocalArtifactProcessor,
    DbtCloudArtifactProcessor,
    ParentRunMetadata,
    UnsupportedDbtCommand,
]



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Optional

import attr
from openlineage.common.utils import get_from_nullable_chain

from great_expectations.core import ExpectationValidationResult


@attr.s
class ExpectationsParserResult:
    """
    Internal class to represent actual expectation values, per table and optionally per column
    """

    facet_key: str = attr.ib()
    value: Any = attr.ib()
    column_id: Optional[str] = attr.ib(default=None)


@attr.s
class GreatExpectationsAssertion:
    expectationType: str = attr.ib()
    success: bool = attr.ib()
    column: Optional[str] = attr.ib(default=None)


class ExpectationsParser:
    """
    Base expectation parser. Dispatches parser looking at expectation type.
    Implementations should extract result from result dictionary.
    """

    expectation_key: str = ""
    facet_key: str = ""

    @classmethod
    def can_accept(cls, expectation_result: ExpectationValidationResult) -> Optional[Any]:
        expectation_type = get_from_nullable_chain(
            expectation_result, ["expectation_config", "expectation_type"]
        )
        return expectation_type and expectation_type == cls.expectation_key

    @staticmethod
    def parse_expectation_result(expectation_result: ExpectationValidationResult) -> ExpectationsParserResult:
        raise NotImplementedError("")


class BetweenRowCountExpectationsParser(ExpectationsParser):
    expectation_key = "expect_table_row_count_to_be_between"
    facet_key = "rowCount"

    @classmethod
    def parse_expectation_result(
        cls, expectation_result: ExpectationValidationResult
    ) -> ExpectationsParserResult:
        return ExpectationsParserResult(
            cls.facet_key,
            get_from_nullable_chain(expectation_result, ["result", "observed_value"]),
        )


class EqualRowCountExpectationsParser(BetweenRowCountExpectationsParser):
    expectation_key = "expect_table_row_count_to_equal"


class FileSizeExpectationsParser(ExpectationsParser):
    expectation_key = "expect_file_size_to_be_between"

    @staticmethod
    def parse_expectation_result(expectation_result: dict) -> ExpectationsParserResult:  # type: ignore # noqa
        pass  # TODO: file asset validation


class ColumnExpectationsParser(ExpectationsParser):
    """
    Extractor for column-based expectations. Looks at column name in addition to expectation type
    """

    column = ""

    @classmethod
    def can_accept(cls, expectation_result: ExpectationValidationResult) -> Optional[Any]:
        expectation_type = get_from_nullable_chain(
            expectation_result, ["expectation_config", "expectation_type"]
        )
        extracted_column = get_from_nullable_chain(
            expectation_result, ["expectation_config", "kwargs", "column"]
        )
        return expectation_type and extracted_column and expectation_type == cls.expectation_key


class ValuesNotNullColumnExpectationParser(ColumnExpectationsParser):
    expectation_key = "expect_column_values_to_not_be_null"

    @staticmethod
    def parse_expectation_result(expectation_result: ExpectationValidationResult) -> Any:
        return ExpectationsParserResult(
            "nullCount",
            get_from_nullable_chain(expectation_result, ["result", "unexpected_count"]),
            get_from_nullable_chain(expectation_result, ["expectation_config", "kwargs", "column"]),
        )


class ValuesDistinctExpectationParser(ColumnExpectationsParser):
    expectation_key = "expect_column_unique_value_count_to_be_between"

    @staticmethod
    def parse_expectation_result(expectation_result: ExpectationValidationResult) -> ExpectationsParserResult:
        return ExpectationsParserResult(
            "distinctCount",
            get_from_nullable_chain(expectation_result, ["result", "observed_value"]),
            get_from_nullable_chain(expectation_result, ["expectation_config", "kwargs", "column"]),
        )


class ValuesSumExpectationParser(ColumnExpectationsParser):
    expectation_key = "expect_column_sum_to_be_between"

    @staticmethod
    def parse_expectation_result(expectation_result: ExpectationValidationResult) -> ExpectationsParserResult:
        sum = get_from_nullable_chain(expectation_result, ["result", "observed_value"])
        return ExpectationsParserResult(
            "sum",
            sum,
            get_from_nullable_chain(expectation_result, ["expectation_config", "kwargs", "column"]),
        )


class ValuesCountExpectationParser(ColumnExpectationsParser):
    expectation_key = "expect_column_sum_to_be_between"

    @staticmethod
    def parse_expectation_result(expectation_result: ExpectationValidationResult) -> ExpectationsParserResult:
        count = get_from_nullable_chain(expectation_result, ["result", "element_count"])
        return ExpectationsParserResult(
            "count",
            count,
            get_from_nullable_chain(expectation_result, ["expectation_config", "kwargs", "column"]),
        )


class ValuesMaxExpectationParser(ColumnExpectationsParser):
    expectation_key = "expect_column_max_to_be_between"

    @staticmethod
    def parse_expectation_result(expectation_result: ExpectationValidationResult) -> ExpectationsParserResult:
        return ExpectationsParserResult(
            "max",
            get_from_nullable_chain(expectation_result, ["result", "observed_value"]),
            get_from_nullable_chain(expectation_result, ["expectation_config", "kwargs", "column"]),
        )


class ValuesMinExpectationParser(ColumnExpectationsParser):
    expectation_key = "expect_column_min_to_be_between"

    @staticmethod
    def parse_expectation_result(expectation_result: ExpectationValidationResult) -> ExpectationsParserResult:
        return ExpectationsParserResult(
            "min",
            get_from_nullable_chain(expectation_result, ["result", "observed_value"]),
            get_from_nullable_chain(expectation_result, ["expectation_config", "kwargs", "column"]),
        )


class ValuesQuantileExpectationParser(ColumnExpectationsParser):
    expectation_key = "expect_column_quantile_values_to_be_between"

    @staticmethod
    def quantile_to_map(observations):
        return {str(k): v for k, v in zip(observations["quantiles"], observations["values"])}

    @classmethod
    def parse_expectation_result(cls, expectation_result: dict) -> ExpectationsParserResult:
        observed_values = get_from_nullable_chain(expectation_result, ["result", "observed_value"])
        return ExpectationsParserResult(
            "quantiles",
            cls.quantile_to_map(observed_values) if observed_values else None,
            get_from_nullable_chain(expectation_result, ["expectation_config", "kwargs", "column"]),
        )


EXPECTATIONS_PARSERS = [
    BetweenRowCountExpectationsParser,
    EqualRowCountExpectationsParser,
    # FileSizeExpectationsParser,
]

COLUMN_EXPECTATIONS_PARSER = [
    ValuesNotNullColumnExpectationParser,
    ValuesDistinctExpectationParser,
    ValuesMinExpectationParser,
    ValuesMaxExpectationParser,
    ValuesQuantileExpectationParser,
    ValuesSumExpectationParser,
    ValuesCountExpectationParser,
]



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import traceback
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

import attr
from openlineage.client.facet import (
    BaseFacet,
    ExternalQueryRunFacet,
    OutputStatisticsOutputDatasetFacet,
)
from openlineage.common.dataset import Dataset, Source
from openlineage.common.models import DbColumn, DbTableSchema
from openlineage.common.schema import GITHUB_LOCATION
from openlineage.common.sql import DbTableMeta
from openlineage.common.utils import get_from_nullable_chain

_BIGQUERY_CONN_URL = "bigquery"

# we lazy-load bigquery Client in BigQueryDatasetsProvider if not type-checking
if TYPE_CHECKING:
    from google.cloud.bigquery import Client


def get_bq_client():
    from google.cloud.bigquery import Client

    return Client


@attr.s
class BigQueryErrorRunFacet(BaseFacet):
    """
    Represents errors that can happen during execution of BigqueryExtractor
    :param clientError: represents errors originating in bigquery client
    :param parserError: represents errors that happened during parsing SQL provided to bigquery
    """

    clientError: str = attr.ib(default=None)
    parserError: str = attr.ib(default=None)

    @staticmethod
    def _get_schema() -> str:
        return GITHUB_LOCATION + "bq-error-run-facet.json"


@attr.s
class BigQueryJobRunFacet(BaseFacet):
    """
    Facet that represents relevant statistics of bigquery run.
    :param cached: bigquery caches query results. Rest of the statistics will not be provided
        for cached queries.
    :param billedBytes: how many bytes bigquery bills for.
    :param properties: full property tree of bigquery run.
    """

    cached: bool = attr.ib()
    billedBytes: Optional[int] = attr.ib(default=None)
    properties: Optional[str] = attr.ib(default=None)

    @staticmethod
    def _get_schema() -> str:
        return GITHUB_LOCATION + "bq-statistics-run-facet.json"


@attr.s
class BigQueryStatisticsDatasetFacet(BaseFacet):
    """
    Facet that represents statistics of output dataset resulting from bigquery run.
    :param outputRows: how many rows query produced.
    :param size: size of output dataset in bytes.
    """

    rowCount: int = attr.ib()
    size: int = attr.ib()

    def to_openlineage(self) -> OutputStatisticsOutputDatasetFacet:
        return OutputStatisticsOutputDatasetFacet(rowCount=self.rowCount, size=self.size)

    @staticmethod
    def _get_schema() -> str:
        return GITHUB_LOCATION + "bq-statistics-dataset-facet.json"


@attr.s
class BigQueryFacets:
    run_facets: Dict[str, BaseFacet] = attr.ib()
    inputs: List[Dataset] = attr.ib()
    output: Optional[Dataset] = attr.ib(default=None)


class BigQueryDatasetsProvider:
    def __init__(self, client: Optional["Client"] = None, logger: Optional[logging.Logger] = None):
        if client is None:
            # lazy-load bigquery client since its slow to import (primarily due to pandas)
            self.client = get_bq_client()
        else:
            self.client = client
        if logger is None:
            self.logger: logging.Logger = logging.getLogger(__name__)
        else:
            self.logger = logger

    def get_facets(self, job_id: str) -> BigQueryFacets:
        inputs = []
        output = None
        run_facets = {}
        try:
            try:
                job = self.client.get_job(job_id=job_id)  # type: ignore
                props = job._properties

                run_stat_facet, dataset_stat_facet = self._get_output_statistics(props)

                run_facets.update(
                    {
                        "bigQuery_job": run_stat_facet,
                        "externalQuery": ExternalQueryRunFacet(externalQueryId=job_id, source="bigquery"),
                    }
                )
                inputs = self._get_input_from_bq(props)
                output = self._get_output_from_bq(props)
                if output and dataset_stat_facet:
                    output.custom_facets.update({"stats": dataset_stat_facet})
                    output.output_facets.update({"outputStatistics": dataset_stat_facet.to_openlineage()})

            finally:
                # Ensure client has close() defined, otherwise ignore.
                # NOTE: close() was introduced in python-bigquery v1.23.0
                if hasattr(self.client, "close"):
                    self.client.close()
        except Exception as e:
            self.logger.error(f"Cannot retrieve job details from BigQuery.Client. {e}", exc_info=True)
            run_facets.update(
                {
                    "bigQuery_error": BigQueryErrorRunFacet(
                        clientError=f"{e}: {traceback.format_exc()}",
                    )
                }
            )
        return BigQueryFacets(run_facets, inputs, output)

    def _get_output_statistics(
        self, properties
    ) -> Tuple[BigQueryJobRunFacet, Optional[BigQueryStatisticsDatasetFacet]]:
        stages = get_from_nullable_chain(properties, ["statistics", "query", "queryPlan"])
        json_props = json.dumps(properties)

        if not stages:
            if get_from_nullable_chain(properties, ["statistics", "query", "statementType"]) in [
                "CREATE_VIEW",
                "CREATE_TABLE",
                "ALTER_TABLE",
            ]:
                return BigQueryJobRunFacet(cached=False), None

            # we're probably getting cached results
            if get_from_nullable_chain(properties, ["statistics", "query", "cacheHit"]):
                return BigQueryJobRunFacet(cached=True), None
            if get_from_nullable_chain(properties, ["status", "state"]) != "DONE":
                raise ValueError("Trying to extract data from running bigquery job")
            raise ValueError(f"BigQuery properties did not have required data: queryPlan - {json_props}")

        out_stage = stages[-1]
        out_rows = out_stage.get("recordsWritten", None)
        out_bytes = out_stage.get("shuffleOutputBytes", None)
        billed_bytes = get_from_nullable_chain(properties, ["statistics", "query", "totalBytesBilled"])
        return BigQueryJobRunFacet(
            cached=False,
            billedBytes=int(billed_bytes) if billed_bytes else None,
            properties=json_props,
        ), BigQueryStatisticsDatasetFacet(
            rowCount=int(out_rows), size=int(out_bytes)
        ) if out_bytes and out_rows else None

    def _get_input_from_bq(self, properties):
        bq_input_tables = get_from_nullable_chain(properties, ["statistics", "query", "referencedTables"])
        if not bq_input_tables:
            return []

        input_table_names = [self._bq_table_name(bq_t) for bq_t in bq_input_tables]
        sources = [self._source() for bq_t in bq_input_tables]
        try:
            return [
                Dataset.from_table_schema(source=source, table_schema=table_schema)
                for table_schema, source in zip(self._get_table_schemas(input_table_names), sources)
            ]
        except Exception as e:
            self.logger.warning(f"Could not extract schema from bigquery. {e}")
            return [Dataset.from_table(source, table) for table, source in zip(input_table_names, sources)]

    def _get_output_from_bq(self, properties) -> Optional[Dataset]:
        bq_output_table = get_from_nullable_chain(properties, ["configuration", "query", "destinationTable"])
        if not bq_output_table:
            return None

        output_table_name = self._bq_table_name(bq_output_table)
        source = self._source()

        table_schema = self._get_table_safely(output_table_name)
        if table_schema:
            return Dataset.from_table_schema(
                source=source,
                table_schema=table_schema,
            )
        else:
            self.logger.warning("Could not resolve output table from bq")
            return Dataset.from_table(source, output_table_name)

    def _get_table_safely(self, output_table_name):
        try:
            return self._get_table(output_table_name)
        except Exception as e:
            self.logger.warning(f"Could not extract output schema from bigquery. {e}")
        return None

    def _get_table_schemas(self, tables: List[str]) -> List[DbTableSchema]:
        # Avoid querying BigQuery by returning an empty array
        # if no tables have been provided.
        if not tables:
            return []

        return [self._get_table(table) for table in tables]

    def _get_table(self, table: str) -> Optional[DbTableSchema]:
        bq_table = self.client.get_table(table)
        if not bq_table._properties:
            return None
        table_prop = bq_table._properties

        fields = get_from_nullable_chain(table_prop, ["schema", "fields"])
        if not fields:
            return None

        columns = [
            DbColumn(
                name=fields[i].get("name"),
                type=fields[i].get("type"),
                description=fields[i].get("description"),
                ordinal_position=i,
            )
            for i in range(len(fields))
        ]

        return DbTableSchema(
            schema_name=table_prop.get("tableReference").get("projectId")
            + "."
            + table_prop.get("tableReference").get("datasetId"),
            table_name=DbTableMeta(table_prop.get("tableReference").get("tableId")),
            columns=columns,
        )

    def _source(self) -> Source:
        return Source(scheme="bigquery", connection_url="bigquery")

    def _bq_table_name(self, bq_table):
        project = bq_table.get("projectId")
        dataset = bq_table.get("datasetId")
        table = bq_table.get("tableId")
        return f"{project}.{dataset}.{table}"



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import logging
import traceback
from typing import Any, Dict, List, Optional

import attr
import botocore
from openlineage.client.facet import (
    BaseFacet,
    ErrorMessageRunFacet,
    OutputStatisticsOutputDatasetFacet,
)
from openlineage.common.dataset import Dataset, Source
from openlineage.common.models import DbColumn, DbTableSchema
from openlineage.common.sql import DbTableMeta


@attr.s
class RedshiftFacets:
    run_facets: Dict[str, BaseFacet] = attr.ib()
    inputs: List[Dataset] = attr.ib()
    output: List[Dataset] = attr.ib()


class RedshiftDataDatasetsProvider:
    def __init__(
        self,
        client: botocore.client,
        connection_details: Dict[str, Any],
        logger: Optional[logging.Logger] = None,
    ):
        if logger is None:
            self.logger: logging.Logger = logging.getLogger(__name__)
        else:
            self.logger = logger
        self.connection_details = connection_details
        self.client = client

    def get_facets(
        self, job_id: str, inputs: List[DbTableMeta], outputs: List[DbTableMeta]
    ) -> RedshiftFacets:
        ds_inputs = []
        ds_outputs = []
        run_facets: Dict[str, BaseFacet] = {}
        dataset_stat_facet = None

        source = Source(
            scheme="redshift",
            authority=self._get_authority(),
        )
        try:
            properties = self.client.describe_statement(Id=job_id)
            dataset_stat_facet = self._get_output_statistics(properties)
        except Exception as e:
            self.logger.error(
                f"Cannot retrieve job details from Redshift Data client. {e}",
                exc_info=True,
            )
            run_facets.update(
                {
                    "errorMessage": ErrorMessageRunFacet(
                        message=f"Cannot retrieve job details from Redshift Data client. {e}",
                        programmingLanguage="PYTHON",
                        stackTrace=f"{e}: {traceback.format_exc()}",
                    )
                }
            )

        ds_inputs = self._get_dataset_from_tables(inputs, source)
        ds_outputs = self._get_dataset_from_tables(outputs, source)

        for ds_output in ds_outputs:
            ds_output.custom_facets.update({"stats": dataset_stat_facet})

        return RedshiftFacets(run_facets, ds_inputs, ds_outputs)

    def _get_output_statistics(self, properties) -> OutputStatisticsOutputDatasetFacet:
        return OutputStatisticsOutputDatasetFacet(
            rowCount=properties.get("ResultRows"),
            size=properties.get("ResultSize"),
        )

    def _get_dataset_from_tables(self, tables: List[DbTableMeta], source: Source) -> Dataset:
        try:
            return [
                Dataset.from_table_schema(
                    source=source,
                    table_schema=table_schema,
                    database_name=self.connection_details.get("database"),
                )
                for table_schema in self._get_table_schemas(tables)
            ]
        except Exception as e:
            self.logger.warning(f"Could not extract schema from redshift. {e}")
            return [Dataset.from_table(source, table.name) for table in tables]

    def _get_authority(self) -> str:
        return (
            f"{self.connection_details['cluster_identifier']}."
            f"{self.connection_details.get('region')}:5439"
        )

    def _get_table_safely(self, output_table_name):
        try:
            return self._get_table(output_table_name)
        except Exception as e:
            self.logger.warning(f"Could not extract output schema from redshift. {e}")
        return None

    def _get_table_schemas(self, tables: List[DbTableMeta]) -> List[DbTableSchema]:
        if not tables:
            return []
        return [self._get_table(table) for table in tables]

    def _get_table(self, table: DbTableMeta) -> Optional[DbTableSchema]:
        kwargs: Dict[str, Any] = {
            "ClusterIdentifier": self.connection_details.get("cluster_identifier"),
            "Database": table.database or self.connection_details.get("database"),
            "Table": table.name,
            "DbUser": self.connection_details.get("db_user"),
            "SecretArn": self.connection_details.get("secret_arn"),
            "Schema": table.schema,
        }
        filter_values = {key: val for key, val in kwargs.items() if val is not None}
        redshift_table = self.client.describe_table(**filter_values)
        fields = redshift_table.get("ColumnList")
        if not fields:
            return None
        schema_name = fields[0]["schemaName"]
        columns = [
            DbColumn(
                name=fields[i].get("name"),
                type=fields[i].get("typeName"),
                ordinal_position=i,
            )
            for i in range(len(fields))
        ]

        return DbTableSchema(
            schema_name=schema_name,
            table_name=DbTableMeta(redshift_table.get("TableName")),
            columns=columns,
        )



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from urllib.parse import quote, urlparse, urlunparse


def fix_account_name(name: str) -> str:
    spl = name.split(".")
    if len(spl) == 1:
        account = spl[0]
        region, cloud = "us-west-1", "aws"
    elif len(spl) == 2:
        account, region = spl
        cloud = "aws"
    else:
        account, region, cloud = spl
    return f"{account}.{region}.{cloud}"


def fix_snowflake_sqlalchemy_uri(uri: str) -> str:
    """Snowflake sqlalchemy connection URI has following structure:
    'snowflake://<user_login_name>:<password>@<account_identifier>/<database_name>/<schema_name>?warehouse=<warehouse_name>&role=<role_name>'
    We want to canonicalize account identifier. It can have two forms:
    - newer, in form of <organization>-<id>. In this case we want to do nothing.
    - older, composed of <id>-<region>-<cloud> where region and cloud can be
      optional in some cases.If <cloud> is omitted, it's AWS.
      If region and cloud are omitted, it's AWS us-west-1
    """

    try:
        parts = urlparse(uri)
    except ValueError:
        # snowflake.sqlalchemy.URL does not quote `[` and `]`
        # that's a rare case so we can run more debugging code here
        # to make sure we replace only password
        parts = urlparse(uri.replace("[", quote("[")).replace("]", quote("]")))

    hostname = parts.hostname
    if not hostname:
        return uri

    # old account identifier like xy123456
    if "." in hostname or not any(word in hostname for word in ["-", "_"]):
        hostname = fix_account_name(hostname)
    # else - its new hostname, just return it
    return urlunparse((parts.scheme, hostname, parts.path, parts.params, parts.query, parts.fragment))



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, ClassVar, List, Optional

from openlineage.client.utils import RedactMixin

if TYPE_CHECKING:
    from openlineage.common.sql import DbTableMeta


class DbColumn(RedactMixin):
    _skip_redact: ClassVar[List[str]] = ["name", "type", "ordinal_position"]

    def __init__(
        self,
        name: str,
        type: str,
        description: Optional[str] = None,
        ordinal_position: Optional[int] = None,
    ):
        self.name = name
        self.type = type
        self.description = description
        self.ordinal_position = ordinal_position

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.type == other.type
            and self.description == other.description
            and self.ordinal_position == other.ordinal_position
        )

    def __repr__(self):
        return f"DbColumn({self.name!r},{self.type!r}, \
                          {self.description!r},{self.ordinal_position!r})"


class DbTableSchema(RedactMixin):
    _skip_redact: ClassVar[List[str]] = ["schema_name", "table_name"]

    def __init__(
        self,
        schema_name: str,
        table_name: "DbTableMeta",
        columns: List[DbColumn],
    ):
        self.schema_name = schema_name
        self.table_name = table_name
        self.columns = columns

    def __eq__(self, other):
        return (
            self.schema_name == other.schema_name
            and self.table_name == other.table_name
            and self.columns == other.columns
        )

    def __repr__(self):
        return f"DbTableSchema({self.schema_name!r},{self.table_name!r}, \
                               {self.columns!r})"



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
from typing import Any, Dict, List, Optional, Tuple, TypeVar

import yaml
from jinja2 import Environment, Undefined
from openlineage.common.provider.dbt.processor import DbtArtifactProcessor
from openlineage.common.utils import get_from_nullable_chain

DBT_TARGET_PATH_ENVVAR = "DBT_TARGET_PATH"
DEFAULT_TARGET_PATH = "target"


T = TypeVar("T")


class LazyJinjaLoadDict(dict):
    """
    A dictionary that lazily renders Jinja2 templates in its values.

    This class is useful for passing data to templates without having to pre-render all of the data.
    It works by traversing the dictionary and rendering any string value using Jinja2.
    If the value is a dictionary, a new `LazyJinjaLoadDict` instance is created and returned.
    """

    def __init__(self, *args, jinja_env, **kwargs):
        super().__init__(*args, **kwargs)
        self.jinja_env = jinja_env

    def __getitem__(self, item):
        arg = dict.__getitem__(self, item)
        return LazyJinjaLoadDict.render_values_jinja(self.jinja_env, arg)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    @staticmethod
    def render_values_jinja(jinja_env, value: T):
        """
        Traverses passed dictionary and render any string value using jinja.

        Returns lazy load dictionary when value is dictionary instance.
        """
        if isinstance(value, list):
            return [LazyJinjaLoadDict.render_values_jinja(jinja_env, elem) for elem in value]
        elif isinstance(value, str):
            return jinja_env.from_string(value).render()
        elif isinstance(value, dict):
            return LazyJinjaLoadDict(value, jinja_env=jinja_env)
        else:
            return value


class SkipUndefined(Undefined):
    def __getattr__(self, name):
        return SkipUndefined(name=f"{self._undefined_name}.{name}")

    def __str__(self):
        return f"{{{{ {self._undefined_name} }}}}"

    def _fail_with_undefined_error(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        arguments = ", ".join(
            [arg._undefined_name if isinstance(arg, SkipUndefined) else str(arg) for arg in args]
        )
        return f"{{{{ {self._undefined_name}({arguments}) }}}}"


class DbtLocalArtifactProcessor(DbtArtifactProcessor):
    should_raise_on_unsupported_command = True

    def __init__(
        self,
        project_dir: str,
        profile_name: Optional[str] = None,
        target: Optional[str] = None,
        target_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.jinja_environment: Optional[Environment] = None

        absolute_dir = os.path.abspath(project_dir)
        dbt_project = self.load_yaml_with_jinja(os.path.join(project_dir, "dbt_project.yml"))
        self.target_path = target_path
        target_path = self.build_target_path(dbt_project)

        self.manifest_path = os.path.join(absolute_dir, target_path, "manifest.json")
        self.run_result_path = os.path.join(absolute_dir, target_path, "run_results.json")
        self.catalog_path = os.path.join(absolute_dir, target_path, "catalog.json")

        self.target = target
        self.project_name = dbt_project["name"]
        self.profile_name = profile_name or dbt_project.get("profile")
        if not self.profile_name:
            raise KeyError(f"profile not found in {dbt_project}")

    def build_target_path(self, dbt_project: dict, target_path: Optional[str] = None) -> str:
        """
        Build dbt target path. Uses the following:
        1. target_path (user-defined value, normally given in --target-path CLI flag)
        2. DBT_TARGET_PATH environment variable
        3. target-path in dbt_project.yml
        4. default ("target")

        Precedence order: user-defined target_path > env var > dbt_project.yml > default

        Reference:
        https://docs.getdbt.com/reference/project-configs/target-path
        """
        return (
            self.target_path
            or os.getenv(DBT_TARGET_PATH_ENVVAR)
            or dbt_project.get("target-path")
            or DEFAULT_TARGET_PATH
        )

    @classmethod
    def load_metadata(
        cls, path: str, desired_schema_versions: List[int], logger: logging.Logger
    ) -> Dict[Any, Any]:
        with open(path) as f:
            metadata = json.load(f)
            str_schema_version = get_from_nullable_chain(metadata, ["metadata", "dbt_schema_version"])
            schema_version = cls.get_schema_version(metadata)
            if schema_version not in desired_schema_versions:
                if schema_version > max(desired_schema_versions):
                    logger.warning(
                        f"Artifact schema version: {str_schema_version} is above dbt-ol "
                        f"supported version {max(desired_schema_versions)}. "
                        f"This might cause errors."
                    )
                else:
                    raise ValueError(
                        f"Wrong version of dbt metadata: {schema_version}, "
                        f"should be in {desired_schema_versions}"
                    )
            return metadata

    @staticmethod
    def env_var(var: str, default: Optional[str] = None) -> str:
        """The env_var() function. Return the environment variable named 'var'.
        If there is no such environment variable set, return the default.

        If the default is None, raise an exception for an undefined variable.
        """
        if var in os.environ:
            return os.environ[var]
        elif default is not None:
            return default
        else:
            msg = f"Env var required but not provided: '{var}'"
            raise Exception(msg)

    @staticmethod
    def load_yaml(path: str) -> Dict:
        with open(path) as f:
            return yaml.safe_load(f)

    @staticmethod
    def setup_jinja() -> Environment:
        env = Environment(extensions=["jinja2.ext.do"], undefined=SkipUndefined)
        # When using env vars for Redshift port, it must be "{{ env_var('PORT') | as_number }}"
        # otherwise Redshift driver will complain, hence the need to add the "as_number" filter
        env.filters.update({"as_number": lambda x: x})
        env.globals["env_var"] = DbtLocalArtifactProcessor.env_var
        return env

    def load_yaml_with_jinja(self, path: str) -> Dict:
        loaded = self.load_yaml(path)
        if not self.jinja_environment:
            self.jinja_environment = self.setup_jinja()
        return LazyJinjaLoadDict(loaded, jinja_env=self.jinja_environment)

    def get_dbt_metadata(
        self,
    ) -> Tuple[Dict[Any, Any], Dict[Any, Any], Dict[Any, Any], Optional[Dict[Any, Any]]]:
        manifest = self.load_metadata(self.manifest_path, [2, 3, 4, 5, 6, 7], self.logger)

        run_result = self.load_metadata(self.run_result_path, [2, 3, 4, 5], self.logger)

        try:
            catalog: Optional[Dict[Any, Any]] = self.load_metadata(self.catalog_path, [1], self.logger)
        except FileNotFoundError:
            catalog = None

        profile_dir = run_result["args"]["profiles_dir"]

        profile = self.load_yaml_with_jinja(os.path.join(profile_dir, "profiles.yml"))[self.profile_name]

        if not self.target:
            self.target = profile["target"]

        profile = profile["outputs"][self.target]

        return manifest, run_result, profile, catalog



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import pytest
from openlineage.client.facet import (
    DataSourceDatasetFacet,
    SchemaDatasetFacet,
    SchemaField,
)
from openlineage.client.run import Dataset as OpenLineageDataset
from openlineage.common.dataset import Dataset, Source
from openlineage.common.models import DbColumn, DbTableSchema
from openlineage.common.sql import DbTableMeta


@pytest.fixture()
def source():
    return Source(
        scheme="dummy",
        authority="localhost:1234",
        connection_url="dummy://localhost:1234",
    )


@pytest.fixture()
def table_schema(source):
    schema_name = "public"
    table_name = DbTableMeta("discounts")
    columns = [
        DbColumn(name="id", type="int4", ordinal_position=1),
        DbColumn(name="amount_off", type="int4", ordinal_position=2),
        DbColumn(name="customer_email", type="varchar", ordinal_position=3),
        DbColumn(name="starts_on", type="timestamp", ordinal_position=4),
        DbColumn(name="ends_on", type="timestamp", ordinal_position=5),
    ]
    schema = DbTableSchema(schema_name=schema_name, table_name=table_name, columns=columns)
    return source, columns, schema


def test_dataset_from(source):
    dataset = Dataset.from_table(source, "source_table", "public")
    assert dataset == Dataset(source=source, name="public.source_table")


def test_dataset_with_db_name(source):
    dataset = Dataset.from_table(source, "source_table", "public", "food_delivery")
    assert dataset == Dataset(source=source, name="food_delivery.public.source_table")


def test_dataset_to_openlineage(table_schema):
    source_name = "dummy://localhost:1234"
    source, columns, schema = table_schema

    dataset_schema = Dataset.from_table_schema(source, schema)
    assert dataset_schema.to_openlineage_dataset() == OpenLineageDataset(
        namespace=source_name,
        name="public.discounts",
        facets={
            "dataSource": DataSourceDatasetFacet(name=source_name, uri=source_name),
            "schema": SchemaDatasetFacet(
                fields=[
                    SchemaField(name="id", type="int4"),
                    SchemaField(name="amount_off", type="int4"),
                    SchemaField(name="customer_email", type="varchar"),
                    SchemaField(name="starts_on", type="timestamp"),
                    SchemaField(name="ends_on", type="timestamp"),
                ]
            ),
        },
    )



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import List, Optional, Union

from openlineage_sql import (
    ColumnLineage,  # noqa: F401
    ColumnMeta,  # noqa: F401
    DbTableMeta,  # noqa: F401
    ExtractionError,  # noqa: F401
    SqlMeta,
    provider,  # noqa: F401
)
from openlineage_sql import parse as parse_sql

log = logging.getLogger(__name__)


def parse(
    sql: Union[List[str], str],
    dialect: Optional[str] = None,
    default_schema: Optional[str] = None,
) -> Optional[SqlMeta]:
    if isinstance(sql, str):
        sql = [sql]
    try:
        return parse_sql(sql, dialect=dialect, default_schema=default_schema)
    except Exception as e:
        log.error(f"SQL parser failed: {e}")
        return None



# Copyright 2018-2023 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from openlineage.common.sql import DbTableMeta


def test_eq_table_name():
    assert DbTableMeta("public.discounts") == DbTableMeta("public.discounts")
    assert DbTableMeta("discounts") != DbTableMeta("public.discounts")
    assert DbTableMeta("discounts").qualified_name == "discounts"
    assert DbTableMeta("public.discounts").qualified_name == "public.discounts"
